%
% Elastic wave extrapolation IR via elastic 1W operator
%
% Factorization of the full elastic isotropic wave equation
% (for displacement components)
%
% Musa Maharramov, Stanford, May 2012
% MP version, fully-scripted Nov 2012
%
clc;
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% initialization
%
wmin=str2num(getenv('OWE_WMIN'));
wmax=str2num(getenv('OWE_WMAX'));
dw=str2num(getenv('OWE_DW'));
%
nx=str2num(getenv('OWE_NX'));
ny=str2num(getenv('OWE_NY'));
nz=str2num(getenv('OWE_NZ'));
%
% don't tolerate more than 1% amplitude growth
amptol=log(1.01)/nz;
%
dx=str2num(getenv('OWE_DX'));
dy=str2num(getenv('OWE_DY'));
dz=str2num(getenv('OWE_DZ'));
%
MAXITER=str2num(getenv('OWE_MAXITER'));
%
LOGNAME=[getenv('OWE_LOGNAME') '_'...
  getenv('HOSTNAME') '_'...
   num2str(feature('getpid')) '.log'];
%
mu=str2num(getenv('OWE_MU'));
lambda=str2num(getenv('OWE_LAMBDA'));
rho=str2num(getenv('OWE_RHO'));
%
dispx=str2num(getenv('OWE_DISPX'));
dispy=str2num(getenv('OWE_DISPY'));
dispz=str2num(getenv('OWE_DISPZ'));
%
a=lambda+mu;
c1=(mu+a)^.5;
%
% log file
flog = fopen(LOGNAME, 'w');
fprintf(flog, 'One-way Elastic Extrapolation\n');
fprintf(flog, '==============================\n');
sdate=datestr(now);
fprintf(flog, 'START %s\n',sdate);

fprintf(flog, 'nx=%d ny=%d nz=%d\n',nx, ny, nz);
fprintf(flog, 'dx=%g dy=%g dz=%g\n', dx, dy, dz);
fprintf(flog, 'wmin=%g wmax=%g dw=%g\n',...
    wmin, wmax, dw);
fprintf(flog, 'lambda=%g mu=%g rho=%g\n', ...
   lambda, mu, rho);
fprintf(flog, 'dispx=%g dispy=%g dispz=%g\n', ...
   dispx, dispy, dispz);
fprintf(flog, '==============================\n');
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% computation
%
Arphase_spectrum=zeros(3,nx,ny);
Aiphase_spectrum=zeros(3,nx,ny);
Brphase_spectrum=zeros(3,nx,ny);
Biphase_spectrum=zeros(3,nx,ny);
%
% extrapolation operator storage
AEXT=0i*zeros(3,3,nx,ny);
%
% fields
image=0i*zeros(3,nz,nx,ny);
pressure=zeros(nz,nx,ny);
cu=zeros(nx,ny);
cu2=cu;
%
% depth-dependent enhancement
tpow=0.5;
%
% horizontal wavenumbers
kx=zeros(nx,1);
ky=zeros(ny,1);
% mapping to natural order
ix2=zeros(nx,1);
iy2=zeros(ny,1);
%
% wavenumbers
for i=1:nx/2+1
  kx(i)=2*pi*(i-1)/dx/nx;
  ix2(i)=i+nx/2-1;
end
for i=nx/2+2:nx
  kx(i)=2*pi*(i-1-nx)/dx/nx;
  ix2(i)=i-nx/2-1;
end
for i=1:ny/2+1
  ky(i)=2*pi*(i-1)/dy/ny;    
  iy2(i)=i+ny/2-1;
end
for i=ny/2+2:ny
  ky(i)=2*pi*(i-1-ny)/dy/ny;
  iy2(i)=i-ny/2-1;
end
%
% parameters of the source
a1=dispx;a2=dispy;a3=dispz;
d1=5*dx;d2=5*dy;
x0=(nx/2-1)*dx; y0=(ny/2-1)*dy;z0=3*dz;
%
DZZ=dz*eye(3);
%
wmin=wmin*2*pi;wmax=wmax*2*pi;dw=dw*2*pi;
w=wmin;
u=0i*zeros(3,nx,ny);
%us=0i*zeros(3,nx,ny);
% frequency loop
while (w<=wmax)    
  c2=w*rho^.5;
  % initial conditions - injecting a source
  for ix=1:nx
    for iy=1:ny
      u(1,ix,iy)=a1*exp(-(((ix-1)*dx-x0)/d1)^2)*exp(-(((iy-1)*dy-y0)/d2)^2);      
      u(2,ix,iy)=a2*exp(-(((ix-1)*dx-x0)/d1)^2)*exp(-(((iy-1)*dy-y0)/d2)^2);      
      u(3,ix,iy)=a3*exp(-(((ix-1)*dx-x0)/d1)^2)*exp(-(((iy-1)*dy-y0)/d2)^2);      
    end
  end
  % compute the operators
  %
  % initial passband estimate
  % for the current frequency
  PB=(pi/dx)^2+(pi/dy)^2;
  for iy=1:ny
    for ix=1:nx
      P=zeros(3,3);S=P;
      P(1,1)=-(mu+a)*kx(ix)*kx(ix)-mu*ky(iy)*ky(iy);
      P(1,2)=-a*kx(ix)*ky(iy);

      P(2,1)=-a*kx(ix)*ky(iy);
      P(2,2)=-(mu+a)*ky(iy)*ky(iy)-mu*kx(ix)*kx(ix);

      P(3,3)=-mu*(kx(ix)*kx(ix)+ky(iy)*ky(iy));

      S(1,3)=1i*a*kx(ix);
      S(2,3)=1i*a*ky(iy);
      S(3,1)=1i*a*kx(ix);
      S(3,2)=1i*a*ky(iy);
      %
      E=zeros(3,3);
      E(1,1)=mu^.5;
      E(2,2)=mu^.5;
      E(3,3)=c1;
      %
      EINV=zeros(3,3);
      EINV(1,1)=1./E(1,1);
      EINV(2,2)=1./E(2,2);
      EINV(3,3)=1./E(3,3);
      %
      S=S-2.*c2*E;
      %
      % for zero wn extrapolators are =I
      if ((ix==1) && (iy==1))
        AEXT(:,:,ix,iy)=expm(-DZZ*EINV*c2*eye(3,3));
        Arphase_spectrum(:,ix2(ix),iy2(iy))=-DZZ*c2*diag(EINV);
        Aiphase_spectrum(:,ix2(ix),iy2(iy))=0;
      %
        Brphase_spectrum(:,ix2(ix),iy2(iy))=-DZZ*c2*diag(EINV);
        Biphase_spectrum(:,ix2(ix),iy2(iy))=0;
        continue;
      end
      %
      % initial approx
      A1=sqrtm(P);
      %
      X=zeros(9,1);
      rr=zeros(9,1);

      iter=0;
      R=P;
      % solve SEP148 eq (8) using steepest descent
      while norm(R)>1.e-3*norm(P) && iter<MAXITER
        iter=iter+1;
        %
        L1=EINV*(S-A1*E);
        L2=A1*EINV+c2*EINV;
        %
        R=P-A1*L1-c2*(A1+L1);
        F=zeros(9,9);
        %    
        for i=1:3
          for j=1:3
            F(i+(j-1)*3,i+(j-1)*3)=F(i+(j-1)*3,i+(j-1)*3)+c2;
            for k=1:3
              F(i+(j-1)*3,i+(k-1)*3)=F(i+(j-1)*3,i+(k-1)*3)+L1(k,j);
              F(i+(j-1)*3,k+(j-1)*3)=F(i+(j-1)*3,k+(j-1)*3)-L2(i,k)*E(j,j);
            end
            rr(i+(j-1)*3)=R(i,j);
          end
        end  
        %
        X=F\rr;
        %
        for i=1:3
          for j=1:3
            DeltaA1(i,j)=X(i+(j-1)*3);
          end
        end
        %
        % update the horizontal op in the first factor
        A1=A1+DeltaA1;
        %    
      end
      if iter==MAXITER
        disp(['wn '...
   num2str(ix) ',' num2str(iy) ' failed to converge']);
        fprintf (flog,'%s\n',['wn '...
   num2str(ix) ',' num2str(iy) ' failed to converge']);
      end
      %
      A2=EINV*(S-A1*E);
      %
      % eigenvalues of the log extrapolator
      ei1=eig(-DZZ*EINV*(A1+c2*eye(3,3)));
      ei2=eig(-DZZ*EINV*(A2+c2*eye(3,3)));
      % pick the factor with one "stable" and two "unstable"
      % eigenvalues to avoid limiting useful bandwidth;
      %
      % two unstable eigenvalues have the same passband,
      % and the stable one has a smaller passband   
      %
      if pick_branch(ei1,ei2)==2
        OPER=A2; ei=ei2; eii=ei1;
      else
        OPER=A1; ei=ei1; eii=ei2;
      end
      % 
      if  (real(ei(1))>amptol) || (real(ei(2))>amptol)...
              || (real(ei(3))>amptol)...
           && (iter<MAXITER) 
        % shrink the 'common' passband (if the approximation's good)
        PB=min(PB,(kx(ix)^2+ky(iy)^2));
      end
      %
      % compute and save the extrapolator
      AEXT(:,:,ix,iy)=expm(-DZZ*EINV*(OPER+c2*eye(3,3)));
      % factor-operator spectra for output;
      Arphase_spectrum(:,ix2(ix),iy2(iy))=sort(real(ei(:)));
      Aiphase_spectrum(:,ix2(ix),iy2(iy))=sort(imag(ei(:)));
      Brphase_spectrum(:,ix2(ix),iy2(iy))=sort(real(eii(:)));
      Biphase_spectrum(:,ix2(ix),iy2(iy))=sort(imag(eii(:)));
      %
    end
  end  
  for iy=1:ny
    for ix=1:nx
      if ((kx(ix)^2+ky(iy)^2)>=PB)
%        Arphase_spectrum(:,ix2(ix),iy2(iy))=0;      
%        Aiphase_spectrum(:,ix2(ix),iy2(iy))=0;      
      end
    end
  end
  %
  % fft
  for i=1:3
    for iy=1:ny
      for ix=1:nx
        cu(ix,iy)=u(i,ix,iy);
      end
    end
    cu2=fft2(cu);
    for iy=1:ny
      for ix=1:nx
        u(i,ix,iy)=cu2(ix,iy);
      end
    end
  end
  %
  % depth loop
  for iz=1:nz
    disp(['depth step ' num2str(iz)])
    fprintf (flog,'%s\n',['depth step ' num2str(iz)]);
    % imaging condition
    for i=1:3
      for iy=1:ny
        for ix=1:nx
          image(i,iz,ix,iy)=image(i,iz,ix,iy)+...
               (iz^tpow)*u(i,ix,iy);
        end
      end
    end
    for iy=1:ny
      for ix=1:nx
        if ((kx(ix)^2+ky(iy)^2)>=PB)          
          u(:,ix,iy)=0;
        else
          u(:,ix,iy)=AEXT(:,:,ix,iy)*u(:,ix,iy);
        end        
      end
    end
  end 
  disp (['completed frequency ' num2str(w)]);
  fprintf (flog,'%s\n',['completed frequency ' num2str(w)]);
  %
  w=w+dw;
end
%
% inverse fft on the whole image
  for iz=1:nz
    for i=1:3
      for iy=1:ny
        for ix=1:nx
          cu(ix,iy)=image(i,iz,ix,iy);
        end
      end
      cu2=ifft2(cu);
      for iy=1:ny
        for ix=1:nx
          image(i,iz,ix,iy)=cu2(ix,iy);
        end
      end
    end
  end
%
  fprintf(flog, 'Saving image\n');

% save the image to file
  fname=[num2str(wmin) '_' num2str(wmax) 'img.bin']
  fid = fopen(fname, 'w');
  rimage=real(image);
  fwrite(fid, rimage, 'double');
  fclose(fid);

  fprintf(flog, 'Saving spectra\n');

% save the spectra to files
  fname=[num2str(wmin) '_' num2str(wmax) 'Arspec.bin']
  fid = fopen(fname, 'w');
  fwrite(fid, Arphase_spectrum, 'double');
  fclose(fid);
%
  fname=[num2str(wmin) '_' num2str(wmax) 'Aispec.bin']
  fid = fopen(fname, 'w');
  fwrite(fid, Aiphase_spectrum, 'double');
  fclose(fid);
%  
  fname=[num2str(wmin) '_' num2str(wmax) 'Brspec.bin']
  fid = fopen(fname, 'w');
  fwrite(fid, Brphase_spectrum, 'double');
  fclose(fid);
%
  fname=[num2str(wmin) '_' num2str(wmax) 'Bispec.bin']
  fid = fopen(fname, 'w');
  fwrite(fid, Biphase_spectrum, 'double');
  fclose(fid);
%   
  sdate=datestr(now);
  fprintf(flog, 'STOP %s\n',sdate);
  fprintf(flog, '==============================\n');
  fclose(flog);
  exit

